# Copyright 2019 Karsten Roth and Biagio Brattoli
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


#################### LIBRARIES ########################
import warnings
warnings.filterwarnings("ignore")

import os, sys, numpy as np, argparse, imp, datetime, time, pickle as pkl, random, json
os.chdir(os.path.dirname(os.path.realpath(__file__)))

import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt

from tqdm import tqdm
import pandas as pd

import torch, torch.nn as nn
import auxiliaries as aux
import datasets as data

import netlib as netlib
import losses as losses
import evaluate as eval
from scipy.spatial import distance

from attack_lib import FastGradientSignUntargeted,RepresentationAdv
from scipy.stats import norm
import seaborn as sns
import torch.multiprocessing
torch.multiprocessing.set_sharing_strategy('file_system')



################### INPUT ARGUMENTS ###################
parser = argparse.ArgumentParser()

####### Main Parameter: Dataset to use for Training
parser.add_argument('--dataset',      default='cub200',   type=str, help='Dataset to use.')

### General Training Parameters
parser.add_argument('--lr',                default=0.00001,  type=float, help='Learning Rate for network parameters.')
parser.add_argument('--fc_lr_mul',         default=0,        type=float, help='OPTIONAL: Multiply the embedding layer learning rate by this value. If set to 0, the embedding layer shares the same learning rate.')
parser.add_argument('--n_epochs',          default=70,       type=int,   help='Number of training epochs.')
parser.add_argument('--kernels',           default=8,        type=int,   help='Number of workers for pytorch dataloader.')
parser.add_argument('--bs',                default=128 ,     type=int,   help='Mini-Batchsize to use.')
parser.add_argument('--bs_val',                default=60 ,     type=int,   help='Mini-Batchsize to use.')
parser.add_argument('--samples_per_class', default=8,        type=int,   help='Number of samples in one class drawn before choosing the next class. Set to >1 for losses other than ProxyNCA.')
parser.add_argument('--seed',              default=1,        type=int,   help='Random seed for reproducibility.')
parser.add_argument('--scheduler',         default='step',   type=str,   help='Type of learning rate scheduling. Currently: step & exp.')
parser.add_argument('--gamma',             default=0.3,      type=float, help='Learning rate reduction after tau epochs.')
parser.add_argument('--decay',             default=0.0004,   type=float, help='Weight decay for optimizer.')
parser.add_argument('--tau',               default=[30,55],nargs='+',type=int,help='Stepsize(s) before reducing learning rate.')

##### Loss-specific Settings
parser.add_argument('--loss',         default='marginloss', type=str,   help='loss options: marginloss, triplet, npair, proxynca')
parser.add_argument('--sampling',     default='distance',   type=str,   help='For triplet-based losses: Modes of Sampling: random, semihard, distance.')
### MarginLoss
parser.add_argument('--margin',       default=0.2,          type=float, help='TRIPLET/MARGIN: Margin for Triplet-based Losses')
parser.add_argument('--beta_lr',      default=0.0005,       type=float, help='MARGIN: Learning Rate for class margin parameters in MarginLoss')
parser.add_argument('--beta',         default=1.2,          type=float, help='MARGIN: Initial Class Margin Parameter in Margin Loss')
parser.add_argument('--nu',           default=0,            type=float, help='MARGIN: Regularisation value on betas in Margin Loss.')
parser.add_argument('--beta_constant',                      action='store_true', help='MARGIN: Use constant, un-trained beta.')
### ProxyNCA
parser.add_argument('--proxy_lr',     default=0.00001,     type=float, help='PROXYNCA: Learning Rate for Proxies in ProxyNCALoss.')
### NPair L2 Penalty
parser.add_argument('--l2npair',      default=0.02,        type=float, help='NPAIR: Penalty-value for non-normalized N-PAIR embeddings.')

##### Evaluation Settings
parser.add_argument('--k_vals',       nargs='+', default=[1,2,4,8], type=int, help='Recall @ Values.')

##### Network parameters
parser.add_argument('--embed_dim',    default=128,         type=int,   help='Embedding dimensionality of the network. Note: in literature, dim=128 is used for ResNet50 and dim=512 for GoogLeNet.')
parser.add_argument('--arch',         default='resnet50',  type=str,   help='Network backend choice: resnet50, googlenet.')
parser.add_argument('--not_pretrained',                    action='store_true', help='If added, the network will be trained WITHOUT ImageNet-pretrained weights.')
parser.add_argument('--grad_measure',                      action='store_true', help='If added, gradients passed from embedding layer to the last conv-layer are stored in each iteration.')
parser.add_argument('--dist_measure',                      action='store_true', help='If added, the ratio between intra- and interclass distances is stored after each epoch.')

##### Setup Parameters
parser.add_argument('--gpu',          default=0,           type=int,   help='GPU-id for GPU to use.')
parser.add_argument('--loadpath',     default='Training_Results/cub200/resnet_cub_margin_dist/',          type=str,   help='load checkpoit')

parser.add_argument('--source_path',  default='.',         type=str, help='Path to training data.')

#Attacker parameters
parser.add_argument('--adv_test',  default='global',         type=str, help='attack the test data or not')

parser.add_argument('--attack_loss',  default='sim',         type=str, help='attack the test data or not')



parser.add_argument('--epsilon',  default=0.0314,         type=float, help='radius of linf attack')
parser.add_argument('--alpha',  default=0.007,         type=float, help='learning rate of attacker')
parser.add_argument('--max_iters',  default=7,         type=int, help='attack steps')
parser.add_argument('--attack_type',  default='linf',         type=str, help='attack steps')
parser.add_argument('--pretrain_arch',    default='normal', type=str, help='swav pretrain.')
parser.add_argument('--sigma',    default=0.01, type=float, help='sigma')
parser.add_argument('--savename',  default="xD",         type=str, help='radius of linf attack')
# parser.add_argument('--attack_loss',  default='sim',        type=str, help='attack the test data or not')
# parser.add_argument('--attack_loss',  default='sim',         type=str, help='attack the test data or not')

##### Read in parameters
opt = parser.parse_args()

"""============================================================================"""
opt.source_path += '/'+opt.dataset


if opt.dataset=='online_products':
    opt.k_vals = [1,10,100,1000]
if opt.dataset=='in-shop':
    opt.k_vals = [1,10,20,30,50]
if opt.dataset=='vehicle_id':
    opt.k_vals = [1,5]

if opt.loss == 'proxynca':
    opt.samples_per_class = 1
else:
    assert not opt.bs%opt.samples_per_class, 'Batchsize needs to fit number of samples per class for distance sampling and margin/triplet loss!'

if opt.loss == 'npair' or opt.loss == 'proxynca': opt.sampling = 'None'

opt.pretrained = not opt.not_pretrained


"""============================================================================"""
################### GPU SETTINGS ###########################
os.environ["CUDA_DEVICE_ORDER"]   ="PCI_BUS_ID"
os.environ["CUDA_VISIBLE_DEVICES"]= str(opt.gpu)


"""============================================================================"""
#################### SEEDS FOR REPROD. #####################
torch.backends.cudnn.deterministic=True
np.random.seed(opt.seed); random.seed(opt.seed)
torch.manual_seed(opt.seed); torch.cuda.manual_seed(opt.seed); torch.cuda.manual_seed_all(opt.seed)


"""============================================================================"""
##################### NETWORK SETUP ##################
opt.device = torch.device('cuda')
#Depending on the choice opt.arch, networkselect() returns the respective network model
model      = netlib.networkselect(opt)

print('{} Setup for {} with {} sampling on {} complete with #weights: {}'.format(opt.loss.upper(), opt.arch.upper(), opt.sampling.upper(), opt.dataset.upper(), aux.gimme_params(model)))
model.load_state_dict(torch.load(opt.loadpath+'checkpoint.pth.tar')['state_dict'])
#Push to Device
_          = model.to(opt.device)





"""============================================================================"""
#################### DATALOADER SETUPS ##################
#Returns a dictionary containing 'training', 'testing', and 'evaluation' dataloaders.
#The 'testing'-dataloader corresponds to the validation set, and the 'evaluation'-dataloader
#Is simply using the training set, however running under the same rules as 'testing' dataloader,
#i.e. no shuffling and no random cropping.
dataloaders      = data.give_dataloaders(opt.dataset, opt)
#Because the number of supervised classes is dataset dependent, we store them after
#initializing the dataloader
opt.num_classes  = len(dataloaders['training'].dataset.avail_classes)




"""============================================================================"""
#################### CREATE LOGGING FILES ###############
#Each dataset usually has a set of standard metrics to log. aux.metrics_to_examine()
#returns a dict which lists metrics to log for training ('train') and validation/testing ('val')


Rep = RepresentationAdv(model, epsilon=opt.epsilon, alpha=opt.alpha, min_val=-10, max_val=10, max_iters=opt.max_iters, _type=opt.attack_type,num_sample = opt.samples_per_class, loss_type=opt.attack_loss, regularize = 'other')
opt.Rep=Rep

"""============================================================================"""
##################### OPTIONAL EVALUATIONS #####################
#Store the averaged gradients returned from the embedding to the last conv. layer.
# if opt.grad_measure:
#     grad_measure = eval.GradientMeasure(opt, name='baseline')
# #Store the relative distances between average intra- and inter-class distance.
# if opt.dist_measure:
#     #Add a distance measure for training distance ratios
#     distance_measure = eval.DistanceMeasure(dataloaders['evaluation'], opt, name='Train', update_epochs=1)
#     # #If uncommented: Do the same for the test set
#     # distance_measure_test = eval.DistanceMeasure(dataloaders['testing'], opt, name='Train', update_epochs=1)




"""============================================================================"""



"""============================================================================"""



"""============================================================================"""
"""========================== MAIN Attacking PART =============================="""
"""============================================================================"""
################### SCRIPT MAIN ##########################
print('\n-----\n')

### Evaluate
_ = model.eval()
# #Each dataset requires slightly different dataloaders.
# if opt.dataset in ['cars196', 'cub200', 'online_products']:
#     eval_params = {'dataloader':dataloaders['testing'], 'model':model, 'opt':opt, 'epoch':0}
# elif opt.dataset=='in-shop':
#     eval_params = {'query_dataloader':dataloaders['testing_query'], 'gallery_dataloader':dataloaders['testing_gallery'], 'model':model, 'opt':opt, 'epoch':0}
# elif opt.dataset=='vehicle_id':
#     eval_params = {'dataloaders':[dataloaders['testing_set1'], dataloaders['testing_set2'], dataloaders['testing_set3']], 'model':model, 'opt':opt, 'epoch':0}

#Compute Evaluation metrics, print them and store in LOG.
# eval.evaluate(opt.dataset, LOG, save=True, **eval_params)

F1, NMI, recall_all_k, feature_coll = aux.eval_metrics_one_dataset(model, test_dataloader=dataloaders['testing'], device=opt.gpu, k_vals=opt.k_vals, opt=opt)
# print(F1, NMI, recall_all_k)
result_str = ', '.join('@{0}: {1:.4f}'.format(k,rec) for k,rec in zip(opt.k_vals, recall_all_k))
result_str = 'Epoch (Test) {0}: NMI [{1:.4f}] | F1 [{2:.4f}] | Recall [{3}]'.format(0, NMI, F1, result_str)
print(result_str)
feature_coll,target = feature_coll
avail_labels  = np.unique(target)

#Compute indixes of embeddings for each class.
class_pos = {}
class_neg = {}
for lab in avail_labels:
    class_pos[lab] = np.where(target==lab)[0]
    class_neg[lab] = np.where(target!=lab)[0]
    
dists = distance.cdist(feature_coll,feature_coll)
all_d = []
for i in range(dists.shape[0]):
    dists[i,i] = 999
    m = dists[i]
#     print(target[i])
#     print(class_pos[target[i]])
    min_pos_dist = np.min(m[class_pos[target[i][0]]])
    min_neg_dist = np.min(m[class_neg[target[i][0]]])
    d = min_neg_dist - min_pos_dist
    all_d.append(d)
all_d = np.array(all_d)-4*(8*np.log(129/0.04)/3/100000)**0.5
radius = 2*opt.sigma*norm.ppf(0.5+all_d/8)
# print(radius)
# ax=sns.histplot(radius,)
ran = np.linspace(0,2,100)
per = []
for j in ran:
    per.append(sum(radius>j)/len(radius))
sns.set(rc={'figure.figsize':(5,5)},font_scale=1.3,style = "whitegrid")
sns.lineplot(ran,per)
# plt.savefig("histplt/"+opt.savename+".png")
np.save("radius/"+opt.savename+".npy",radius)

print('\n-----\n')
